from __future__ import annotations

import os
from tempfile import TemporaryDirectory
from typing import Any, Dict, Optional

from google.auth.credentials import Credentials
from google.cloud.aiplatform import PipelineJob
from kfp.v2.compiler import Compiler

from unipipe.backend.kfp import KubeflowPipelinesBackend
from unipipe.dsl import Pipeline
from unipipe.executor.base import Executor

# from google.auth import default
# from google.cloud.storage import Client as StorageClient

# class PipelineState(str, Enum):
#     """
#     Enum for possible pipeline states.
#     """

#     running: str = "PIPELINE_STATE_RUNNING"
#     succeeded: str = "PIPELINE_STATE_SUCCEEDED"
#     failed: str = "PIPELINE_STATE_FAILED"
#     cancelled: str = "PIPELINE_STATE_CANCELLED"
#     cancelling: str = "PIPELINE_STATE_CANCELLING"
#     paused: str = "PIPELINE_STATE_PAUSED"
#     pending: str = "PIPELINE_STATE_PENDING"


# class TaskDetail(BaseModel):
#     taskName: Optional[str]
#     state: Optional[str]
#     execution: Optional[Dict]
#     inputs: Optional[Dict]
#     outputs: Optional[Dict]

#     @property
#     def metadata(self) -> str:
#         if self.execution:
#             return self.execution["metadata"]

#         return None

#     @property
#     def execution_state(self) -> str:
#         if self.execution:
#             return self.execution["state"]

#         return None


# class JobDetail(BaseModel):
#     taskDetails: Optional[List[TaskDetail]]


# class PipelineStatus(BaseModel):
#     """
#     Dataclass for pipeline status.
#     """

#     name: str
#     displayName: str
#     createTime: datetime.datetime
#     updateTime: datetime.datetime
#     pipelineSpec: Dict
#     state: PipelineState
#     jobDetail: Optional[JobDetail]

#     @property
#     def task_details(self) -> Optional[List[TaskDetail]]:
#         return self.jobDetail.taskDetails


class VertexExecutor(Executor):
    def run(
        self,
        pipeline: Any,
        arguments: Optional[Dict] = None,
        pipeline_root: Optional[str] = None,
        enable_caching: bool = False,
        credentials: Optional[Credentials] = None,
        project: str = None,
        location: str = "us-central1",
    ):
        if isinstance(pipeline, Pipeline):
            pipeline = KubeflowPipelinesBackend().build(pipeline)

        # StorageClient()
        # default_credentials, default_project = default()
        # if not credentials:
        #     credentials = default_credentials
        # if not project:
        #     project = default_project

        with TemporaryDirectory() as tempdir:
            template_path = os.path.join(tempdir, "pipeline.json")
            Compiler().compile(pipeline_func=pipeline, package_path=template_path)
            PipelineJob(
                display_name="example-vertex-pipeline",
                template_path=template_path,
                parameter_values=arguments,
                credentials=credentials,
                project=project,
                location=location,
                pipeline_root=pipeline_root,
                enable_caching=enable_caching,
            ).submit()



from __future__ import annotations

from typing import Any, Dict

import kfp.dsl as kfp_dsl
import kfp.v2.dsl as kfp_v2_dsl
from kfp.v2.components.component_factory import create_component_from_func

from unipipe.dsl import Component, ConditionalPipeline, LazyAttribute, Pipeline
from unipipe.utils.annotations import resolve_annotations


def build_kubeflow_component(component: Component):
    comp = create_component_from_func(
        func=resolve_annotations(component.func),
        base_image=component.base_image or "fkodom/unipipe:latest",
        packages_to_install=component.packages_to_install,
        pip_index_urls=component.pip_index_urls,
    )
    comp.name = component.name
    return comp


def set_hardware_attributes(container_op: Any, component: Component):
    hardware = component.hardware
    if hardware.cpus:
        container_op.set_cpu_limit(hardware.cpus)
    if hardware.memory:
        container_op.set_memory_limit(hardware.memory)
    if hardware.accelerator:
        accelerator = hardware.accelerator
        if accelerator.count:
            # KFP requires the count to be given as a string
            container_op.set_gpu_limit(str(accelerator.count))
        if accelerator.type:
            container_op.add_node_selector_constraint(
                "cloud.google.com/gke-accelerator",
                accelerator.type.value,
            )

    return container_op


def resolve_value(_locals: Dict, value: Any) -> Any:
    if isinstance(value, LazyAttribute):
        return _locals[value.parent.name].outputs[value.key]
    elif isinstance(value, Component):
        return _locals[value.name].output
    elif isinstance(value, Pipeline):
        if value.name in _locals:
            return _locals[value.name]
        return resolve_value(_locals, value.return_value)
    elif isinstance(value, (tuple, list)):
        return tuple(resolve_value(_locals, x) for x in value)
    else:
        return value


def run_component(component: Component, **kwargs):
    result = component.func(**kwargs)
    kfp_component = build_kubeflow_component(component)
    result = kfp_component(**kwargs)
    set_hardware_attributes(result, component)

    return result


def build_pipeline_with_locals(pipeline: Pipeline, _locals: Dict[str, Any]):
    for component in pipeline.components:
        kwargs = {k: resolve_value(_locals, v) for k, v in component.inputs.items()}
        __locals = {**_locals, **kwargs}
        if isinstance(component, ConditionalPipeline):
            operand1 = resolve_value(_locals, component.condition.operand1)
            operand2 = resolve_value(_locals, component.condition.operand2)
            comparator = component.condition.comparator

            # An unfortunate fact about KFP conditions -- they require two operands,
            # and as a result, they don't deal well with raw boolean input values.
            # As a workaround, you can cast the boolean values to strings and compare.
            #
            # Basically, this allows 'unipipe' users to write things like:
            #   with dsl.equal(evaluates_to_true, True):
            #       do_something()
            if isinstance(operand1, bool):
                operand1 = str(operand1).lower()
            if isinstance(operand2, bool):
                operand2 = str(operand2).lower()

            with kfp_dsl.Condition(comparator(operand1, operand2)):
                result, _locals = build_pipeline_with_locals(
                    component, _locals=__locals
                )
        elif isinstance(component, Pipeline):
            result, _ = build_pipeline_with_locals(component, _locals=__locals)
        elif isinstance(component, Component):
            result = run_component(component, **kwargs)
        else:
            raise TypeError(
                f"Found pipeline component {component} with unexpected type: "
                f"{type(component)}. Valid component types are "
                "[Component, ConditionalPipeline, Pipeline]."
            )

        _locals[component.name] = result

    return_value = resolve_value(_locals, pipeline.return_value)
    return return_value, _locals


class KubeflowPipelinesBackend:
    def build(self, pipeline: Pipeline):
        @kfp_v2_dsl.pipeline(name=pipeline.name)
        def kfp_pipeline():
            build_pipeline_with_locals(pipeline, _locals=pipeline.inputs)

        return kfp_pipeline



import pytest
import torch

import unipipe
from examples.ex01_hello_world import pipeline as pipeline_01
from examples.ex02_hello_pipeline import pipeline as pipeline_02
from examples.ex03_multi_output_components import pipeline as pipeline_03
from examples.ex04_pipeline_arguments import pipeline as pipeline_04
from examples.ex05_dependency_management import pipeline as pipeline_05
from examples.ex06_hardware_specs import pipeline as pipeline_06
from examples.ex07_nested_pipelines import pipeline as pipeline_07
from examples.ex08_control_flow import pipeline as pipeline_08
from examples.ex09_advanced_control_flow import bad_pipeline as bad_pipeline_09
from examples.ex09_advanced_control_flow import good_pipeline as pipeline_09


@pytest.mark.docker
def test_example_01():
    unipipe.run(pipeline=pipeline_01(), executor="docker")


@pytest.mark.docker
def test_example_02():
    unipipe.run(pipeline=pipeline_02(), executor="docker")


@pytest.mark.docker
def test_example_03():
    unipipe.run(pipeline=pipeline_03(), executor="docker")


@pytest.mark.docker
def test_example_04():
    unipipe.run(pipeline=pipeline_04(name="Tyrion Lannister"), executor="docker")


@pytest.mark.docker
def test_example_05():
    unipipe.run(pipeline=pipeline_05(name="Tyrion Lannister"), executor="docker")


@pytest.mark.docker
def test_example_06():
    if torch.cuda.is_available():
        image_url = "https://raw.githubusercontent.com/EliSchwartz/imagenet-sample-images/master/n01443537_goldfish.JPEG"
        unipipe.run(pipeline=pipeline_06(image_url=image_url), executor="docker")


@pytest.mark.docker
def test_example_07():
    unipipe.run(pipeline=pipeline_07(), executor="docker")


@pytest.mark.docker
def test_example_08():
    unipipe.run(pipeline=pipeline_08(), executor="docker")


@pytest.mark.docker
def test_example_09():
    with pytest.raises(KeyError):
        unipipe.run(pipeline=bad_pipeline_09(name="Ned Stark"), executor="docker")
    unipipe.run(pipeline=bad_pipeline_09(name="Tyrion Lannister"), executor="docker")
    unipipe.run(pipeline=pipeline_09(name="Ned Stark"), executor="docker")



from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional

from unipipe.utils.annotations import get_annotations, infer_input_types


class MultipleDispatch:
    def __init__(self):
        self.funcs: List[Callable] = []
        self.signatures: List[Dict] = []

    def add(self, func: Callable, signature: Optional[Dict] = None):
        if not signature:
            signature = get_annotations(func, eval_str=True)
        assert signature is not None
        if "return" in signature:
            signature.pop("return")
        self.funcs.append(func)
        self.signatures.append(signature)

    def __getitem__(self, inputs: Dict[str, Any]):
        signature = infer_input_types(inputs)
        for func, _signature in zip(self.funcs, self.signatures):
            if not len(signature) == len(_signature):
                continue
            elif all(issubclass(v, _signature[k]) for k, v in signature.items()):
                return func

        raise TypeError(
            f"Could not find a function with 'signature={signature}'. Available "
            f"signatures include: [{self.signatures}]."
        )

    def __call__(self, **inputs):
        func = self[inputs]
        return func(**inputs)


class DispatchRegistry:
    d: Dict[str, MultipleDispatch] = defaultdict(MultipleDispatch)

    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super().__new__(cls)
        return cls.instance

    def __getitem__(self, name: str) -> MultipleDispatch:
        return self.d[name]

    def __call__(self, func: Callable, signature: Optional[Dict] = None):
        name = func.__qualname__
        self[name].add(func, signature=signature)
        return self.d[name]


dispatch = DispatchRegistry()


# fmt: off

@dispatch
def len_(a: str) -> int: return len(a)  # noqa: E704, F811
@dispatch  # type: ignore
def len_(a: tuple) -> int: return len(a)  # noqa: E704, F811


@dispatch
def int_(a: str) -> int: return int(a)  # noqa: E704, F811
@dispatch  # type: ignore
def int_(a: int) -> int: return int(a)  # noqa: E704, F811
@dispatch  # type: ignore
def int_(a: float) -> int: return int(a)  # noqa: E704, F811
@dispatch  # type: ignore
def int_(a: bool) -> int: return int(a)  # noqa: E704, F811


@dispatch
def float_(a: str) -> float: return float(a)  # noqa: E704, F811
@dispatch  # type: ignore
def float_(a: int) -> float: return float(a)  # noqa: E704, F811
@dispatch  # type: ignore
def float_(a: float) -> float: return float(a)  # noqa: E704, F811
@dispatch  # type: ignore
def float_(a: bool) -> float: return float(a)  # noqa: E704, F811


@dispatch
def str_(a: str) -> str: return str(a)  # noqa: E704, F811
@dispatch  # type: ignore
def str_(a: int) -> str: return str(a)  # noqa: E704, F811
@dispatch  # type: ignore
def str_(a: float) -> str: return str(a)  # noqa: E704, F811
@dispatch  # type: ignore
def str_(a: bool) -> str: return str(a)  # noqa: E704, F811


# TODO:
#   - abs
#   - round
#   - ceil
#   - floor
#   - ge
#   - gt
#   - le
#   - lt
#   - ne
#   - contains


@dispatch
def equal(a: str, b: str): return a == b  # noqa: E704, F811
@dispatch  # type: ignore
def equal(a: int, b: int): return a == b  # noqa: E704, F811
@dispatch  # type: ignore
def equal(a: int, b: float): return a == b  # noqa: E704, F811
@dispatch  # type: ignore
def equal(a: float, b: int): return a == b  # noqa: E704, F811
@dispatch  # type: ignore
def equal(a: float, b: float): return a == b  # noqa: E704, F811
@dispatch  # type: ignore
def equal(a: tuple, b: tuple): return a == b  # noqa: E704, F811


@dispatch
def add(a: str, b: str) -> str: return a + b  # noqa: E704, F811
@dispatch  # type: ignore
def add(a: int, b: int) -> int: return a + b  # noqa: E704, F811
@dispatch  # type: ignore
def add(a: int, b: bool) -> int: return a + b  # noqa: E704, F811
@dispatch  # type: ignore
def add(a: int, b: float) -> float: return a + b  # noqa: E704, F811
@dispatch  # type: ignore
def add(a: float, b: int) -> float: return a + b  # noqa: E704, F811
@dispatch  # type: ignore
def add(a: float, b: bool) -> float: return a + b  # noqa: E704, F811
@dispatch  # type: ignore
def add(a: float, b: float) -> float: return a + b  # noqa: E704, F811
@dispatch  # type: ignore
def add(a: bool, b: bool) -> bool: return a + b  # noqa: E704, F811
@dispatch  # type: ignore
def add(a: bool, b: int) -> int: return a + b  # noqa: E704, F811
@dispatch  # type: ignore
def add(a: bool, b: float) -> float: return a + b  # noqa: E704, F811


@dispatch
def sub(a: int, b: int) -> int: return a - b  # noqa: E704, F811
@dispatch  # type: ignore
def sub(a: int, b: bool) -> int: return a - b  # noqa: E704, F811
@dispatch  # type: ignore
def sub(a: int, b: float) -> float: return a - b  # noqa: E704, F811
@dispatch  # type: ignore
def sub(a: float, b: int) -> float: return a - b  # noqa: E704, F811
@dispatch  # type: ignore
def sub(a: float, b: bool) -> float: return a - b  # noqa: E704, F811
@dispatch  # type: ignore
def sub(a: float, b: float) -> float: return a - b  # noqa: E704, F811
@dispatch  # type: ignore
def sub(a: bool, b: bool) -> bool: return a - b  # noqa: E704, F811
@dispatch  # type: ignore
def sub(a: bool, b: int) -> int: return a - b  # noqa: E704, F811
@dispatch  # type: ignore
def sub(a: bool, b: float) -> float: return a - b  # noqa: E704, F811


@dispatch
def mul(a: str, b: int) -> str: return a * b  # noqa: E704, F811
@dispatch  # type: ignore
def mul(a: int, b: str) -> str: return a * b  # noqa: E704, F811
@dispatch  # type: ignore
def mul(a: int, b: int) -> int: return a * b  # noqa: E704, F811
@dispatch  # type: ignore
def mul(a: int, b: bool) -> int: return a * b  # noqa: E704, F811
@dispatch  # type: ignore
def mul(a: int, b: float) -> float: return a * b  # noqa: E704, F811
@dispatch  # type: ignore
def mul(a: float, b: int) -> float: return a * b  # noqa: E704, F811
@dispatch  # type: ignore
def mul(a: float, b: bool) -> float: return a * b  # noqa: E704, F811
@dispatch  # type: ignore
def mul(a: float, b: float) -> float: return a * b  # noqa: E704, F811
@dispatch  # type: ignore
def mul(a: bool, b: bool) -> int: return a * b  # noqa: E704, F811
@dispatch  # type: ignore
def mul(a: bool, b: int) -> int: return a * b  # noqa: E704, F811
@dispatch  # type: ignore
def mul(a: bool, b: float) -> float: return a * b  # noqa: E704, F811


@dispatch
def div(a: int, b: int) -> int: return a / b  # noqa: E704, F811
@dispatch  # type: ignore
def div(a: int, b: float) -> float: return a / b  # noqa: E704, F811
@dispatch  # type: ignore
def div(a: float, b: int) -> float: return a / b  # noqa: E704, F811
@dispatch  # type: ignore
def div(a: float, b: float) -> float: return a / b  # noqa: E704, F811


@dispatch
def floordiv(a: int, b: int) -> int: return a // b  # noqa: E704, F811
@dispatch  # type: ignore
def floordiv(a: int, b: float) -> float: return a // b  # noqa: E704, F811
@dispatch  # type: ignore
def floordiv(a: float, b: int) -> float: return a // b  # noqa: E704, F811
@dispatch  # type: ignore
def floordiv(a: float, b: float) -> float: return a // b  # noqa: E704, F811


@dispatch
def contains(a: str, b: str) -> bool: return a.__contains__(b)  # noqa: E704, F811

# fmt: on



from __future__ import annotations

from inspect import isclass
from typing import Any, Dict

from unipipe.dsl import Component, ConditionalPipeline, LazyAttribute, Pipeline
from unipipe.executor.base import LocalExecutor
from unipipe.utils.annotations import get_annotations


class PythonExecutor(LocalExecutor):
    def resolve_local_value(self, _locals: Dict, value: Any) -> Any:
        if isinstance(value, LazyAttribute):
            return getattr(self.resolve_local_value(_locals, value.parent), value.key)
        elif isinstance(value, Pipeline):
            if value.name in _locals:
                return _locals[value.name]
            return self.resolve_local_value(_locals, value.return_value)
        elif isinstance(value, (tuple, list)):
            return tuple(self.resolve_local_value(_locals, x) for x in value)
        elif isinstance(value, Component):
            assert isinstance(_locals, dict)
            return _locals[value.name]
        else:
            return value

    def run_component(self, component: Component, **kwargs):
        result = component.func(**kwargs)
        return_type = get_annotations(component.func, eval_str=True).get("return")
        if isclass(return_type) and issubclass(return_type, tuple):
            result = return_type(*result)

        return result

    def run_conditional_pipeline_with_locals(
        self, pipeline: ConditionalPipeline, _locals: Dict[str, Any]
    ):
        operand1 = self.resolve_local_value(_locals, pipeline.condition.operand1)
        operand2 = self.resolve_local_value(_locals, pipeline.condition.operand2)
        comparator = pipeline.condition.comparator

        if comparator(operand1, operand2):
            return self.run_pipeline_with_locals(pipeline, _locals=_locals)
        else:
            return None, _locals
